/* array.c - functions for handling dynamic arrays
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: array.c 366 2004-03-14 17:27:24Z rsnel $
 */
#include <config.h>
#include <string.h> /* memmove, strcpy */
#include <stdlib.h> /* realloc, free */
#include <stdio.h>  /* snprintf */
#include <assert.h>
#include "exception.h"
#include "array.h"

#define SANE ((*d->objects && *d->no_objects > 0) || \
	(!*d->objects && *d->no_objects == 0)) && d->size > 0
#define OBJECT(a) *d->objects + (a)*d->size

void la_array_init(la_array_t *d, void **objects, int *no_objects, 
		int size, void (*freer)()) {
	assert(d);
	assert(size > 0);
	assert(objects);
	d->objects = objects; 
	d->no_objects =  no_objects;
	d->size = size;
	d->freer = freer;
	d->strerr[0] = '\0';
}

void *la_array_get(la_array_t *d, int i) {
	assert(d);
	assert(SANE);
	assert(i >= 0 && i < *d->no_objects);
	return d->objects + d->size * i;
}

void *la_array_insert(la_array_t *d, void *object, int i) {
	void *objects;
	assert(d);
	assert(SANE);
	assert(i >= 0 && i <= *d->no_objects);
	objects = realloc(*d->objects, d->size*(*d->no_objects + 1));
	if (!objects) {
		snprintf(d->strerr, LA_ARRAY_STRERR_LEN, 
				"out of memory (re) allocating %d bytes",
				d->size*(*d->no_objects + 1));
		throw();
	}
	*d->objects = objects;
	if (i != *d->no_objects) 
		memmove(OBJECT(i+1), OBJECT(i), d->size*(*d->no_objects - i));
	(*d->no_objects)++;
	if (object) memcpy(OBJECT(i), object, d->size);
	else memset(OBJECT(i), 0, d->size);
	return OBJECT(i);
}

void la_array_delete(la_array_t *d, void *object, int i) {
	void *objects;
	assert(d);
	assert(SANE);
	assert(i >= 0 && i < *d->no_objects);
	if (object) memmove(object, OBJECT(i), d->size);
	if (d->freer) (*d->freer)(OBJECT(i));
	(*d->no_objects)--;
	if (i != *d->no_objects)
		memmove(OBJECT(i), OBJECT(i + 1), d->size*(*d->no_objects - i));
	else if (*d->no_objects == 0) {
		free(*d->objects);
		*d->objects = NULL;
		return;
	}
	objects = realloc(*d->objects, d->size*(*d->no_objects));
	/* shrink to nonzero should imho never lead to errors
	 * in realloc, at least the old pointer will stay valid, and
	 * obviously the corresponding memory area large enough */
	if (objects) *d->objects = objects;
}

void la_array_free(la_array_t *d) {
	int i;
	assert(d);
	if (*d->objects == NULL && *d->no_objects == 0 
			&& d->freer == NULL && d->size == 0) return;
	assert(SANE);
	if (d->freer) for (i = 0; i < *d->no_objects; i++)
		(*d->freer)(OBJECT(i));
	free(*d->objects);
}

void la_array_empty(la_array_t *d) {
	assert(d);
	assert(SANE);
	la_array_free(d);
	*d->objects = NULL;
	*d->no_objects = 0;
}

