/* crypt.c - trivial stream PRNG XOR based 'cypher'
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: crypt.c 365 2004-03-14 17:19:16Z rsnel $
 */
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "../verbose.h"
#include "../char_io.h"

typedef struct {
	int c1, c2, c3;
	int state;
} random_t;

typedef struct {
	random_t r;
	int active;
} priv_crypt_t;

/* Generic deterministic pseudo random number generator,
 * it generates numbers 0, 1, ..., c3. (if c1 and c2 are chosen
 * carefully. For an explanation of simple deterministic pseudo random 
 * number generators, search google for simple PRNG. */
static void rnd_next(random_t *r) {
	r->state = (r->state*r->c1 + r->c2)&r->c3;
}

#define LA_VERBOSE_TYPE Char_IO

static int init(void **priv, void *argument) {
	priv_crypt_t *p;

	p = malloc(sizeof(*p));
	if (!p) {
		LA_ERROR("%s", strerror(errno));
		return LA_CHAR_IO_ERROR;
	}
	
	p->r.c1 = 0x80FD43FD;
	p->r.c2 = 0x00C39EC3;
	p->r.c3 = 0x00FFFFFF;
	p->r.state = 0x0A09E86;
	p->active = 1;
	*priv = p;
	return LA_CHAR_IO_OK;
}

static int filter(void *priv, char *next_in, int *avail_in,
		char *next_out, int *avail_out, int flush, char *error) {
	priv_crypt_t *p = priv;
	int rnd;

	if (flush) assert(*avail_in == 0);

	while (*avail_in > 0 && *avail_out > 0) { 
		rnd_next(&p->r);
		if (p->active) {
			rnd = 255*((double)p->r.state/(double)(p->r.c3+1));
			*next_out++ = rnd^*next_in++;	
		} else *next_out++ = *next_in++;
		(*avail_in)--; (*avail_out)--;
	}

	if (flush) return LA_CHAR_IO_FLUSHED;

	return LA_CHAR_IO_OK;
}

static int control(void *priv, char *command, void *param) {
	priv_crypt_t *p = priv;
	if (!strcmp(command, "active")) {
		p->active = 1;
		return LA_CHAR_IO_OK;
	}
	if (!strcmp(command, "passive")) {
		p->active = 0;
		return LA_CHAR_IO_OK;
	}
	if (!strcmp(command, "getXOR")) {
		*(int*)param = 255*((double)p->r.state/(double)(p->r.c3+1));
		return LA_CHAR_IO_OK;
	}
	LA_ERROR("unkown command %s", command);
	return LA_CHAR_IO_ERROR;
}

static void uninit(void *priv) {
	free(priv);
}

la_char_filter_info_t la_char_filter_crypt = { &init, &filter, &control, 
	&uninit, "crypt" };

