/* char_io.c - functions for handling character input/output
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: char_io.c 456 2004-04-03 13:50:14Z rsnel $
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include "array.h"
#include "exception.h"
#include "emem.h"
#include "verbose.h"
#include "char_io.h"
#include "adrift.h"

static la_char_filter_info_t *la_char_filter_infos[LA_CHAR_FILTERS] = { 
	&la_char_filter_crypt,
	&la_char_filter_crlf2z,
	&la_char_filter_zescape,
	&la_char_filter_linebreak,
	&la_char_filter_zunescape,
	&la_char_filter_fpout,
	&la_char_filter_fout,
	&la_char_filter_inflate,
	&la_char_filter_deflate,
	&la_char_filter_tagkill,
	&la_char_filter_bufout,
	&la_char_filter_null,
	&la_char_filter_tee,
	&la_char_filter_fpin
};

#define LA_VERBOSE_TYPE	Char_IO

void la_char_io_init(la_char_io_t *io, char *name) {
	assert(name);
	LA_DEBUG("initializing char_io");
	io->no_filters = 0;
	io->filters = NULL;
	io->strerr[0] = '\0';
	io->name = estrdup(name);
	
}

static void char_filter_free(la_char_filter_t *filter) {
	(*filter->filter_info->uninit)(filter->priv);
}

void la_char_io_free(la_char_io_t *io) {
	int i;
	LA_DEBUG("freeing char_io");
	for (i = 0; i < io->no_filters; i++)
		char_filter_free(io->filters + i);

	free(io->filters);
	free(io->name);
}

static la_char_filter_info_t *char_filter_find(char *name) {
	int i;
	for (i = 0; i < LA_CHAR_FILTERS; i++) 
		if (!strcmp(name, la_char_filter_infos[i]->name))
			return la_char_filter_infos[i];
	return NULL;
}

void la_char_filter_push(la_char_io_t *io, char *name, void *argument) {
	la_char_filter_t f;
	la_array_t a;

	LA_DEBUG("attempting to add filter \"%s\" to char_io", name);

	if (!(f.filter_info = char_filter_find(name))) 
		LA_THROW("requested filter \"%s\" not available", name);

	if ((*f.filter_info->init)(&f.priv, argument) != LA_CHAR_IO_OK) throw();

	f.avail_in = 0;
	f.bytecount = 0;

	la_array_sinit(&a, (*io), filters, NULL);
	try la_array_push(&a, &f);
	catch LA_THROW("error pushing filter \"%s\": %s", name, a.strerr);
	endtry;
}

void la_char_filter_pop(la_char_io_t *io) {
	la_array_t a;
	LA_DEBUG("attempting to pop filter from char_io");
	la_array_sinit(&a, (*io), filters, NULL);
	char_filter_free(io->filters + io->no_filters - 1);
	la_array_pop(&a, NULL);
}

void la_char_filter_bytecount(int *count, la_char_io_t *io, int no) {
	if (no > io->no_filters - 1)
		LA_THROW("illegal filter number specified for bytecount");

	*count = io->filters[io->no_filters - no - 1].bytecount;
}

void la_char_filter_command(la_char_io_t *io, int no, char *cmnd, void *arg) {
	if (no > io->no_filters - 1)
		LA_THROW("illegal filter number specified for command");

	if (io->filters[io->no_filters - no - 1].filter_info->control(
			io->filters[io->no_filters - no - 1].priv, cmnd, 
			arg) != LA_CHAR_IO_OK) throw();
}

static int char_filter_getc(la_char_io_t *cin, int n) {
	la_char_filter_t *f;
	char next_out;
	int c, avail_out = 1;
	
	f = cin->filters + n;

	do {
		if (n != 0 && f->avail_in == 0) {
			c = char_filter_getc(cin, n - 1);
			if (c == EOF) return EOF;
			f->next_in = c;
			f->avail_in = 1;
		}
		cin->err = f->filter_info->filter(f->priv, &f->next_in,
				&f->avail_in, &next_out, &avail_out, 0, 
				cin->strerr);
		if (cin->err != LA_CHAR_IO_OK) return EOF;
	} while (avail_out == 1);

	f->bytecount++;
	return (unsigned char)next_out;
}

static int char_io_getc(la_char_io_t *cin, int whine) {
	int c;
	assert(cin->no_filters > 0);
	c = char_filter_getc(cin, cin->no_filters - 1);
	if (c == EOF) {
		if (whine) LA_THROW("unexpected end of input");
		throw();
	}
	return c;
}

int la_char_io_getc_nowhine(la_char_io_t *cin) {
	return char_io_getc(cin, 0);
}

int la_char_io_getc(la_char_io_t *cin) {
	return char_io_getc(cin, 1);
}

void la_char_io_skip(la_char_io_t *cin, int count) {
	assert(count >= 0);
	while (count-- > 0) {
		try la_char_io_getc(cin);
		catch LA_THROW("skipping %d bytes from character input (EOF)", 
				count);
		endtry;
	}
}

void la_char_io_read(char *buf, int count, la_char_io_t *cin) {
	assert(count >= 0);
	while (count-- > 0) {
		try *(buf++) = la_char_io_getc(cin);
		catch LA_THROW("reading %d bytes from character input (EOF)", 
					count);
		endtry;
	}
}

static int char_filter_putc(int c, la_char_io_t *cout, 
		int n) {
	la_char_filter_t *fs;
	char next_out;
	int avail_out = 1;

	fs = cout->filters + n;

	while (fs->avail_in == 1) {
		cout->err = fs->filter_info->filter(fs->priv, &fs->next_in,
				&fs->avail_in, &next_out, 
				&avail_out, 0, cout->strerr);
		if (cout->err != LA_CHAR_IO_OK) return EOF;
		if (avail_out == 0) {
			assert(n);
			char_filter_putc((int)next_out, cout, n - 1);
			avail_out = 1;
		}
	}

	fs->next_in = c;
	fs->avail_in = 1;

	fs->bytecount++;
	return LA_CHAR_IO_OK;
}

void la_char_io_putc(int c, la_char_io_t *cout) {
	assert(cout->no_filters > 0);
	if (char_filter_putc(c, cout, cout->no_filters - 1) == EOF) throw(); 
}

void la_char_io_write(char *buf, int count, la_char_io_t *cout) {
	assert(count >= 0);
	while (count-- > 0) {
		try la_char_io_putc((int)(*buf++), cout);
		catch LA_THROW("writing %d bytes to character output", count);
		endtry;
	}
}

void la_char_io_puts(char *string, la_char_io_t *cout) {
	la_char_io_write(string, strlen(string), cout);
}

void la_char_io_printf(la_char_io_t *cout, char *format, ...) {
	va_list ap;
	try {
		va_start(ap, format);
		la_char_io_vprintf(cout, format, ap);
		va_end(ap);
	}
	catch {
		va_end(ap);
		throw();
	}
	endtry;
}

void la_char_io_vprintf(la_char_io_t *cout, char *format, va_list ap) {
	char *line;
	int length;
	length = vasprintf(&line, format, ap);
	if (length < 0) LA_THROW("out of memory preparing output string");

	try {
		la_char_io_write(line, length, cout);
		free(line);
	}
	catch {
		free(line);
		throw();
	}
	endtry;
}

static int char_filter_flush(la_char_io_t *cout, int n) {
	la_char_filter_t *fs;
	char next_out;
	int err, avail_out = 1;

	fs = cout->filters + n;

	while (fs->avail_in == 1) {
		err = fs->filter_info->filter(fs->priv, &fs->next_in,
				&fs->avail_in, &next_out, &avail_out, 0,
				cout->strerr);
		if (err != LA_CHAR_IO_OK) return EOF;
		if (avail_out == 0) {
			assert(n);
			char_filter_putc((int)next_out, cout, n - 1);
			avail_out = 1;
		}
	}

	
	err = fs->filter_info->filter(fs->priv, &fs->next_in,
				&fs->avail_in, &next_out, &avail_out, 1,
				cout->strerr);
	if (err == LA_CHAR_IO_ERROR) return EOF;

	while (avail_out == 0 || err != LA_CHAR_IO_FLUSHED) {
		assert(n);
		if (avail_out == 0) {
			char_filter_putc((int)next_out, cout, n - 1);
		}
		avail_out = 1;
		err = fs->filter_info->filter(fs->priv, &fs->next_in,
				&fs->avail_in, &next_out, &avail_out, 1,
				cout->strerr);
		if (err == LA_CHAR_IO_ERROR) return EOF;
	}

	if (n != 0) return char_filter_flush(cout, n - 1);
	return LA_CHAR_IO_OK;
}

void la_char_io_flush(la_char_io_t *cout) {
	assert(cout->no_filters > 0);
	if (char_filter_flush(cout, cout->no_filters - 1) != LA_CHAR_IO_OK) 
		throw();
}

