/* intstack.c - functions for handling (out of order) compound integers
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: intstack.c 366 2004-03-14 17:27:24Z rsnel $
 */
#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "array.h"
#include "intstack.h"
#include "verbose.h"

#define LA_VERBOSE_TYPE IntStack

static void eltfree(void *e) {
	la_intstack_elt_t *elt = e;
	if (elt->type == 1) free(elt->cnts.string);
}

void la_intstack_init(la_intstack_t *s) {
	assert(s);
	s->elts = NULL;
	s->no_elts = 0;
	la_array_init(&s->stack, (void*)&s->elts, &s->no_elts,
			sizeof(*s->elts), NULL);
	s->curr = -1;
}

void la_intstack_free(la_intstack_t *s) {
	s->stack.freer = eltfree;
	la_array_free(&s->stack);
}

int la_intstack_avail(la_intstack_t *s) {
	return (s->curr != -1) ? 1 : 0;
}

int la_intstack_avail_string(la_intstack_t *s) {
	return (s->curr != -1 && s->elts[s->curr].type == 1) ? 1 : 0;
}

void la_intstack_push(la_intstack_t *s, int i, char string[]) {
	la_intstack_elt_t elt;
	elt.type = 0;
	elt.cnts.integer = i;
	strncpy(elt.string, string, 255);
	la_array_insert(&s->stack, &elt, s->curr + 1);
}

void la_intstack_push_string(la_intstack_t *s, char *st, char string[]) {
	la_intstack_elt_t elt;
	LA_DBG2("pushing string: %s", st);
	elt.type = 1;
	elt.cnts.string = st;
	strncpy(elt.string, string, 255);
	la_array_insert(&s->stack, &elt, s->curr + 1);
}

void la_intstack_pop(la_intstack_t *s) {
	s->curr++;
}

void la_intstack_update(la_intstack_t *s, int i, char string[]) {
	assert(s->elts[s->curr].type == 0);
	s->elts[s->curr].cnts.integer = i;
	if (string) strncpy(s->elts[s->curr].string, string, 255);
}

int la_intstack_peek(la_intstack_t *s, char string[]) {
	assert(s->curr < s->no_elts);
	assert(s->elts[s->curr].type == 0);
	strncpy(string, s->elts[s->curr].string, 255);
	return s->elts[s->curr].cnts.integer;
}

int la_intstack_read(la_intstack_t *s, char string[]) {
	la_intstack_elt_t elt;
	assert(s->elts[s->curr].type == 0);
	la_array_delete(&s->stack, &elt, s->curr);
	s->curr--;
	strncpy(string, elt.string, 255);
	return elt.cnts.integer;
}

char *la_intstack_read_string(la_intstack_t *s, char string[]) {
	char *line;
	la_intstack_elt_t elt;
	la_array_delete(&s->stack, &elt, s->curr);
	s->curr--;
	strncpy(string, elt.string, 255);
	if (elt.type == 1) {
		LA_DBG2("reading string: %s", elt.cnts.string);
		return elt.cnts.string;
	}
	asprintf(&line, "%d", elt.cnts.integer);
	if (!line) LA_THROW("out of memory");
	return line;
}

