/* line_in.c - functions for handling input delay from .TAF files
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: line_in.c 453 2004-04-01 19:43:43Z rsnel $
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <setjmp.h>
#include <string.h>
#include "verbose.h"
#include "array.h"
#include "cex.h"
#include "intstack.h"
#include "line_in.h"
#include "emem.h"

#define LA_VERBOSE_TYPE Line_In

void la_line_in_init(la_line_in_t *lin, la_char_io_t *cin, int line_no) {
	assert(cin && lin);
	lin->cin = cin;
	lin->spaceint = 0;
	lin->in_line_no = line_no;
	snprintf(lin->line_no, 256, "%d", line_no);
	la_char_filter_push(lin->cin, "crlf2z", &lin->in_line_no);
	la_intstack_init(&lin->intstack);
}

void la_line_in_free(la_line_in_t *lin) {
	la_char_filter_pop(lin->cin);
	la_intstack_free(&lin->intstack);
}

static char *la_line_in_gets(la_line_in_t *f) {
	char *line = NULL, *tmpline;
	int max_size = 0;
	int c, size = 0;
	do {
		while (size < max_size) {
			try c = la_char_io_getc_nowhine(f->cin);
			catch {
				free(line);
				if (size != 0) LA_THROW("end of input before "
						"end of line, at line %s of "
						"input", 
						f->line_no);
				else LA_THROW("end of input while expecting "
						"another input line after "
						"line %s", f->line_no);
			}
			endtry;
			*(line+size) = c;
			size++;
			if (c == 0) {
				tmpline = realloc(line, size);
				assert(tmpline);
				return tmpline;
			}
		}

		max_size *= 2;
		if (max_size == 0) max_size = 80;

		tmpline = realloc(line, max_size);
		if (!tmpline) {
			free(line);
			LA_THROW("%s", strerror(errno));
		}
		line = tmpline;
	} while (1);
}

void la_line_in_space(la_line_in_t *lin) {
	/* the next int from lin->cin will be a spaceint */
	/* this flag will be autoreset to 0 */
	lin->spaceint = 1;
}

char *la_line_in_get_string(la_line_in_t *lin) {
	char *line;
	if (la_intstack_avail(&lin->intstack)) { 
		/* create string from available int */
		line = la_intstack_read_string(&lin->intstack, lin->line_no);
		LA_DBG2("CREASTRING:%s:\"%s\"", lin->line_no, line);
	} else {
		line = la_line_in_gets(lin);
		snprintf(lin->line_no, 256, "%d",lin->in_line_no);
		LA_DBG2("READ:%s:\"%s\"", lin->line_no, line);
	}
	return line;
}

void la_line_in_skip(la_line_in_t *lin) {
	char *line;
	line = la_line_in_get_string(lin);
	free(line);
}

static int la_line_in_parse_int(la_line_in_t *lin, char *line) {
	char *prefix, *postfix;
	int prefixlen, result, out;
	if (lin->spaceint == 0) {
		prefix = "";
		postfix = "";
	} else {
		prefix = " ";
		postfix = " ";
		lin->spaceint = 0;
	}
	assert(lin);
	prefixlen = strlen(prefix);
	if (prefixlen != 0 && strncmp(line, prefix, prefixlen)) {
		LA_WARNING("expected \"%s\" at start of line %s of %s, not found", prefix, lin->cin->name, lin->line_no);
		prefixlen = 0;
	}
	if (strlen(line) == 0) {
		LA_THROW("line %s of %s empty; expected integer", 
				lin->line_no, lin->cin->name);
	}

	if ((*(line + prefixlen) < '0' || *(line + prefixlen) > '9') && 
			*(line + prefixlen) != '-') {
		LA_WARNING("junk found before expected integer on line %s of %s", lin->line_no, lin->cin->name);
	}
	result = sscanf(line + prefixlen, "%d", &out);
	if (result != 1) {
		LA_THROW("expected integer on line %s of %s, but found \"%s\"", 
				lin->line_no, lin->cin->name, line);
	}
	if (*(line+prefixlen) == '-') prefixlen++;
	while (*(line+prefixlen) >= '0' && *(line+prefixlen) <= '9') 
		prefixlen++;
	if (strcmp(line+prefixlen, postfix)) {
		LA_WARNING("expected \"%s\" at end of line %s of %s, not found",
				postfix, lin->line_no, lin->cin->name);
	}
	return out;
}

int la_line_in_get_int(la_line_in_t *lin) {
	char *line;
	int i;
	if (la_intstack_avail(&lin->intstack)) {
		if (!la_intstack_avail_string(&lin->intstack)) {
			if (lin->spaceint) lin->spaceint = 0;
			return la_intstack_read(&lin->intstack, lin->line_no);
		}
		line = la_intstack_read_string(&lin->intstack, lin->line_no);
	} else line = la_line_in_get_string(lin);
	try {
		la_exception_cleanup_add(line, free);
		i = la_line_in_parse_int(lin, line);
	}
	delegate;

	return i;
}

static int scanford(char *string) {
        char *ptr = string;
        if (*ptr == '\0') return 0;
        while (*(ptr + 1) != '\0') {
                if (!strncmp(ptr, "%d", 2)) return 1;
                ptr++;
        }
        return 0;
}

void la_line_in_break_compound(la_line_in_t *lin, int count, int xtra, ...) {
	va_list ap;
	int in, i;
	char tmp[256];
	char tmp2[256];
	char *string;
	in = la_line_in_get_int(lin);
	va_start(ap, xtra);
	for (i = 0; i < count; i++) {
		string = va_arg(ap, char*);
		snprintf(tmp, 255, "%s[%d]", lin->line_no, i);
		if (scanford(string)) snprintf(tmp2, 255, string, xtra);
		else strncpy(tmp2, string, 255);
		la_intstack_push(&lin->intstack, 
				la_cex(tmp2, in), tmp);
	}
	va_end(ap);
	for (i = 0; i < count; i++) la_intstack_pop(&lin->intstack);
}

void la_line_in_push(la_line_in_t *lin, int count) {
	assert(count >= 0);
	while (count-- > 0)
		la_intstack_push_string(&lin->intstack, 
				la_line_in_get_string(lin), lin->line_no);
}

void la_line_in_pop(la_line_in_t *lin, int count) {
	assert(count >= 0);
	while (count-- > 0)
		la_intstack_pop(&lin->intstack);
}

void la_line_in_fakeint(la_line_in_t *lin, int fake) {
	la_intstack_push(&lin->intstack, fake, "default");
	la_intstack_pop(&lin->intstack);
}

void la_line_in_fakestring(la_line_in_t *lin, char *s) {
	la_intstack_push_string(&lin->intstack, estrdup(s), "default");
	la_intstack_pop(&lin->intstack);
}
