/* line_out.c - functions for handling output delay to .TAF files
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: line_out.c 366 2004-03-14 17:27:24Z rsnel $
 */
#define _GNU_SOURCE
#include <stdarg.h>
#include <errno.h>
#include <setjmp.h>
#include <string.h>
#include <stdlib.h>
#include "verbose.h"
#include "array.h"
#include "cex.h"
#include "intstack.h"
#include "line_out.h"
#include "emem.h"

#define LA_VERBOSE_TYPE Line_Out

void la_line_out_init(la_line_out_t *lout, la_char_io_t *cout, int line_no) {
	output_t fout;
	assert(cout && lout);
	fout.allowed_writes = 0; /* unused bogus value */
	fout.crpos = &lout->root;
	lout->cout = cout;
	lout->spaceint = 0;
	lout->blackhole = 0;
	lout->line_no = line_no;
	lout->root.next = NULL;
	lout->root.type = root_e;
	lout->current_out = 0;
	lout->outputs = NULL;
	lout->no_outputs = 0;
	la_array_init(&lout->outstack, (void*)&lout->outputs, 
			&lout->no_outputs, sizeof(*lout->outputs), NULL);
	la_array_push(&lout->outstack, &fout);
}

static delayed_t *get_cdel(la_line_out_t *lout) {
	return lout->outputs[lout->current_out].crpos;
}

static void putint(la_line_out_t *lout, int out, int space) {
	char *prefix = "";
	if (space) {
		if (out >= 0) prefix = " ";
		la_char_io_printf(lout->cout, "%s%d \r\n", prefix, out);
	} else la_char_io_printf(lout->cout, "%d\r\n", out);
}

static void putstring(la_line_out_t *lout, char *string) {
	la_char_io_printf(lout->cout, "%s\r\n", string);
}

static void ccompound(la_line_out_t *lout, la_intstack_t *is, int in) {
	char string[256];
	int previous, space, cont;
	do {
		cont = 0;
		previous = la_intstack_read(is, string);
		if (!strcmp(string, "cstart") || !strcmp(string, "scstart")) assert(0);
		previous += la_cex(string, in);
		if (!la_intstack_avail(is)) assert(0);
		la_intstack_peek(is, string);
		if (!strcmp(string, "cstart") || !strcmp(string, "scstart")) {
			if (!strcmp(string, "scstart")) space = 1;
			else space = 0;
			la_intstack_read(is, string);
			if (la_intstack_avail(is)) { 
				if (space) assert(0);
				in = previous;
				cont = 1;
			} else putint(lout, previous, space);
		} else la_intstack_update(is, previous, NULL);
	} while (cont);
}

void la_line_out_flush(la_line_out_t *lout) {
	delayed_t *d = &lout->root;
	la_intstack_t is;
	la_intstack_init(&is);
	while (lout->root.next) {
		d = lout->root.next;
		switch (d->type) {
			case integer_e:
				if (!la_intstack_avail(&is))
					putint(lout, d->content.integer, 0);
				else ccompound(lout, &is, d->content.integer);
				break;
			case string_e:
				if (!la_intstack_avail(&is))
					putstring(lout, d->content.string);
				else {
					int i;
					char *end;
					i = strtol(d->content.string, &end, 10);
					if (errno == EINVAL || *end != '\0')
						assert(0);
					ccompound(lout, &is, i);
				}
				break;
			case cstart_e:
				la_intstack_push(&is, 0, "cstart");
				la_intstack_pop(&is);
				break;
			case scstart_e:
				if (la_intstack_avail(&is)) assert(0);
				la_intstack_push(&is, 0, "scstart");
				la_intstack_pop(&is);
				break;
			case spaceint_e:
				if (la_intstack_avail(&is)) assert(0);
				putint(lout, d->content.integer, 1);
				break;
			case centry_e:
				la_intstack_push(&is, 0, d->content.string);
				la_intstack_pop(&is);
				break;
			case root_e:
				assert(0);
		}
		lout->root.next = d->next;
		if (d->type == centry_e) free(d->content.string);
		free(d);
	}
	la_intstack_free(&is);
}

void la_line_out_free(la_line_out_t *lout) {
	delayed_t *d = &lout->root, *next = d->next;
	while (next) {
		d = next;
		next = d->next;
		if (d->type == centry_e) free(d->content.string);
		free(d);
	}
	la_array_free(&lout->outstack);
}

void la_line_out_space(la_line_out_t *lout) {
	/* the next int from lin->cin will be a spaceint */
	/* this flag will be autoreset to 0 */
	lout->spaceint = 1;
}

static void prepend_delayed(la_line_out_t *lout, delayed_t *d) {
	delayed_t *tmp;
	tmp = get_cdel(lout)->next;
	get_cdel(lout)->next = d;
	d->next = tmp;
}
	
static void skip_delayed(la_line_out_t *lout, int i) {
	assert(i >= 0);
	while (i--) 
		lout->outputs[lout->current_out].crpos = get_cdel(lout)->next;
}

static void insert_delayed(la_line_out_t *lout, delayed_t *d) {
	prepend_delayed(lout, d);
	skip_delayed(lout, 1);
}

static void record_insert(la_line_out_t *lout) {
	if (lout->no_outputs - 1 == lout->current_out) return;
	if (--lout->outputs[lout->current_out].allowed_writes == 0)
		la_array_delete(&lout->outstack, NULL, lout->current_out);
}

void la_line_out_put_string(la_line_out_t *lout, char *line) {
	delayed_t *d;
	assert(lout && lout->blackhole >= 0 && lout->spaceint == 0);
	if (lout->blackhole > 0) {
		lout->blackhole--;
		return;
	}
	d = emalloc(sizeof(*d));
	d->type = string_e;
	d->content.string = line;
	insert_delayed(lout, d);
	record_insert(lout);
}

void la_line_out_put_int(la_line_out_t *lout, int out) {
	delayed_t *d;
	assert(lout && lout->blackhole >= 0);
	if (lout->blackhole > 0) {
		lout->blackhole--;
		if (lout->spaceint) lout->spaceint = 0;
		return;
	}
	d = emalloc(sizeof(*d));
	if (lout->spaceint) {
		lout->spaceint = 0;
		d->type = spaceint_e;
	} else d->type = integer_e;
	d->content.integer = out;
	insert_delayed(lout, d);
	record_insert(lout);
}

void la_line_out_skip(la_line_out_t *lout) {
	la_line_out_put_int(lout, 0);
}

static int scanford(char *string) {
	char *ptr = string;
	if (*ptr == '\0') return 0;
	while (*(ptr + 1) != '\0') {
		if (!strncmp(ptr, "%d", 2)) return 1;
		ptr++;
	}
	return 0;
}

void la_line_out_make_compound(la_line_out_t *lout, int count, int xtra, ...) {
	delayed_t *d = emalloc(sizeof(*d));
	int i;
	va_list ap;
	char *string;
	char *as;
	assert(count > 0);
	if (lout->spaceint) {
		lout->spaceint = 0;
		d->type = scstart_e;
	} else d->type = cstart_e;
	insert_delayed(lout, d);
	va_start(ap, xtra);
	for (i = 0; i < count; i++) {
		d = emalloc(sizeof(*d));
		d->type = centry_e;
		string = va_arg(ap, char*);
		if (scanford(string)) {
			asprintf(&as, string, xtra);
			d->content.string = as;
		} else {
			d->content.string = estrdup(string);
		}
		prepend_delayed(lout, d);
	}
	va_end(ap);
	skip_delayed(lout, count);
	if (lout->no_outputs - 1 == lout->current_out) return;
	lout->outputs[lout->current_out].allowed_writes += count -1;
}

static void inputarray_split(la_line_out_t *lout, int count) {
	output_t o;
	assert(count > 0);
	if (lout->current_out != lout->no_outputs - 1) {
		assert(count < lout->outputs[lout->current_out].allowed_writes);
		lout->outputs[lout->current_out].allowed_writes -= count;
	}
	o.allowed_writes = count;
	o.crpos = get_cdel(lout);
	if (lout->current_out - 1 >= 0) 
		if (lout->outputs[lout->current_out-1].crpos == o.crpos) {
			lout->outputs[lout->current_out-1].allowed_writes += count;
			return;
		}
	la_array_insert(&lout->outstack, &o, lout->current_out++);
}

void la_line_out_push(la_line_out_t *lout, int count) {
	assert(count >= 0);
	while (lout->current_out != lout->no_outputs - 1 &&
			count >= lout->outputs[lout->current_out].allowed_writes) 
		count -= lout->outputs[lout->current_out++].allowed_writes;
	if (count == 0) return;
	inputarray_split(lout, count);
}

void la_line_out_pop(la_line_out_t *lout, int count) {
	assert(count >= 0);
	lout->current_out--;
	assert(lout->current_out >= 0);
	while (count >= lout->outputs[lout->current_out].allowed_writes) {
		count -= lout->outputs[lout->current_out].allowed_writes;
		if (count != 0) {
			lout->current_out--;
			assert(lout->current_out >= 0);
		}
	}
	if (count == 0) return;
	inputarray_split(lout, count);
}

void la_line_out_blackhole(la_line_out_t *lout) {
	assert(lout && lout->blackhole >= 0);
	lout->blackhole++;
}

