/* resolve.c - functions for resolving objects 
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: resolve.c 464 2004-04-04 23:13:16Z rsnel $
 */
#include "taf/generic.h"
#include "objcache.h"
#include "object.h"
#include "resolve.h"
#include "restriction.h"

#define LA_VERBOSE_TYPE Resolve

#define rcase(a) \
	case la_taf_object_spec_##a##_e: \
		object->realtype = la_taf_object_spec_direct_e; \
		object->spec.direct = oc->elts[object->spec.a].a; \
		break

void la_resolve_object_f(la_taf_t *taf, la_taf_object_spec_t *object) {
	la_objcache_t *oc = &taf->oc;
	la_taf_object_spec_types_t type = object->realtype;
	if (taf->oc.elts == NULL) la_objcache_init(taf);
	switch (type) {
		rcase(chairlike);
		rcase(bedlike);
		rcase(surface);
		rcase(dynamic);
		rcase(container);
		rcase(clothing);
		rcase(stateful);
		default:
			break;
	}
}

void la_resolve_player_roomlocation(la_tas_t *tas, la_array_t *rooms) {
	la_array_push(rooms, &tas->data.player.location.room.spec.room);
}

void la_resolve_character_roomlocation(la_tas_t *tas, la_array_t *rooms, int character) {
	la_taf_person_location_t *loc;
	loc = &CS(character)->location;
	switch (loc->room.realtype) {
		case la_taf_room_spec_none_e:
			break;
		case la_taf_room_spec_room_e:
			la_array_push(rooms, &loc->room.spec.room);
			break;
		default:
			LA_THROW("unable to resolve definite dynamic location, illegal room_spec");
	}
}

void la_resolve_person_roomlocation(la_tas_t *tas, la_array_t *rooms, la_taf_person_spec_t *pers) {
	switch (pers->realtype) {
		case la_taf_person_spec_player_e:
			la_resolve_player_roomlocation(tas, rooms);
			break;
		case la_taf_person_spec_character_e:
			la_resolve_character_roomlocation(tas, rooms, pers->spec.character);
			break;
		default:
			LA_THROW("unable to resolve definite dynamic location, illegal person_spec");
			break;
	}

}

void la_resolve_static_roomlocation(la_tas_t *tas, la_array_t *rooms, la_taf_static_location_t *loc) {
	int i;
	switch (loc->type) {
		case 0: /* nowhere */
			break;
		case 1: /* one place */
			if (loc->spec.one_place.type)
				la_array_push(rooms, &loc->spec.one_place.spec.room);
			break;
		case 2: /* some places */
			for (i = 0; i < NO_ROOMS; i++) 
				if (*((int*)loc->spec.some_places.some_rooms.ptr + i) == 1)
					la_array_push(rooms, &i);
			break;
		case 3: /* all places */
			for (i = 0; i < NO_ROOMS; i++) la_array_push(rooms, &i);
			break;
		case 4: /* part of character (only really in 4.00?) */
			la_resolve_character_roomlocation(tas, rooms, loc->part_of_which_character);
			break;
	}
}

void la_resolve_object_roomlocation(la_tas_t *tas, la_array_t *rooms, int object) {
	if (is_dynamic(OF(object))) la_resolve_dynamic_roomlocation(tas, rooms, &OS(object)->location);
	else la_resolve_static_roomlocation(tas, rooms, &OF(object)->slocation);
}

void la_resolve_dynamic_roomlocation(la_tas_t *tas, la_array_t *rooms, la_taf_definite_dynamic_location_t *loc) {
	switch (loc->room.realtype) {
		case la_taf_room_spec_none_e:
			break;
		case la_taf_room_spec_room_e:
			la_array_push(rooms, &loc->room.spec.room);
			break;
		case la_taf_room_spec_posture_e:
			switch (loc->posture.realtype) {
				case la_taf_dynamic_posture_held_by_e:
					la_resolve_person_roomlocation(tas, rooms, &loc->posture.spec.held_by);
					break;
				case la_taf_dynamic_posture_worn_by_e:
					la_resolve_person_roomlocation(tas, rooms, &loc->posture.spec.worn_by);
					break;
				case la_taf_dynamic_posture_in_e:
					la_resolve_object_roomlocation(tas, rooms, loc->posture.spec.in.spec.direct);
					break;
				case la_taf_dynamic_posture_on_e:
					la_resolve_object_roomlocation(tas, rooms, loc->posture.spec.in.spec.direct);
					break;
				default:
					LA_THROW("unable to resolve definite dynamic location, illegal posture");
			}
			break;
		default:
			LA_THROW("unable to resolve definite dynamic location, illegal room_spec");
	}
}

void la_resolve_characters_in_room(la_tas_t *tas, la_array_t *characters, int room) {
	int i;
	for (i = 0; i < NO_CHARACTERS; i++) {
		if (la_restriction_character_in_room(tas, i, room)) la_array_push(characters, &i);
	}
}

void la_resolve_characters_in_same_room_as_player(la_tas_t *tas, la_array_t *characters) {
	la_resolve_characters_in_room(tas, characters, tas->data.player.location.room.spec.room);
}

void la_resolve_objects_directly_in_same_room_as_player(la_tas_t *tas, la_array_t *objects) {
	int i;
	la_taf_room_spec_t *room;
	for (i = 0; i < NO_OBJECTS; i++) {
		if (!is_dynamic(OF(i))) continue;
		room = &OS(i)->location.room;
		if (room->realtype == la_taf_room_spec_room_e)
			if (room->spec.room == tas->data.player.location.room.spec.room) la_array_push(objects, &i);
	}
}

void la_resolve_all_held_objects(la_tas_t *tas, la_array_t *objects) {
	int i;
	for (i = 0; i < NO_OBJECTS; i++) {
		if (la_restriction_object_held_by_player(tas, i)) la_array_push(objects, &i);
	}
}

void la_resolve_all_worn_objects(la_tas_t *tas, la_array_t *objects) {
	int i;
	for (i = 0; i < NO_OBJECTS; i++) {
		if (la_restriction_object_worn_by_player(tas, i)) la_array_push(objects, &i);
	}
}

char *la_direction_names[12] = { "north", "east", "south", "west",
	"up", "down", "in", "out", "northeast", "southeast", "southwest", "northwest" };

void la_resolve_allowed_directions(la_tas_t *tas, la_array_t *directions) {
	la_taf_room_t *room = RF(tas->data.player.location.room.spec.room);
	if (room->north.type) la_array_push(directions, &la_direction_names[0]);
	if (room->east.type) la_array_push(directions, &la_direction_names[1]);
	if (room->south.type) la_array_push(directions, &la_direction_names[2]);
	if (room->west.type) la_array_push(directions, &la_direction_names[3]);
	if (room->up.type) la_array_push(directions, &la_direction_names[4]);
	if (room->down.type) la_array_push(directions, &la_direction_names[5]);
	if (room->in.type) la_array_push(directions, &la_direction_names[6]);
	if (room->out.type) la_array_push(directions, &la_direction_names[7]);
	if (room->northeast.type) la_array_push(directions, &la_direction_names[8]);
	if (room->southeast.type) la_array_push(directions, &la_direction_names[9]);
	if (room->southwest.type) la_array_push(directions, &la_direction_names[10]);
	if (room->northwest.type) la_array_push(directions, &la_direction_names[11]);
}

void la_resolve_all_objects_in_container(la_tas_t *tas, la_array_t *objects, int i) {
	int j;
	if (!is_container(OF(i))) return;
	for (j = 0; j < NO_OBJECTS; j++)
		if (OS(j)->location.posture.realtype == la_taf_dynamic_posture_in_e)
			if (OS(j)->location.posture.spec.in.spec.direct == i)
				la_array_push(objects, &j);
}

void la_resolve_all_objects_on_surface(la_tas_t *tas, la_array_t *objects, int i) {
	int j;
	if (!is_surface(OF(i))) return;
	for (j = 0; j < NO_OBJECTS; j++)
		if (OS(j)->location.posture.realtype == la_taf_dynamic_posture_on_e)
			if (OS(j)->location.posture.spec.on.spec.direct == i)
				la_array_push(objects, &j);
}

void la_resolve_object(la_tas_t *tas, la_array_t *objects, la_taf_object_spec_t *object) {
	la_taf_object_spec_t o;
	switch (object->realtype) {
		case la_taf_object_spec_the_floor_e:
		case la_taf_object_spec_not_selected_e:
		case la_taf_object_spec_none_e:
			break;
		case la_taf_object_spec_any_e:
			LA_THROW("unable to resolve 'any' object");
		case la_taf_object_spec_chairlike_e:
		case la_taf_object_spec_bedlike_e:
		case la_taf_object_spec_container_e:
		case la_taf_object_spec_surface_e:
		case la_taf_object_spec_stateful_e:
		case la_taf_object_spec_direct_e:
		case la_taf_object_spec_clothing_e:
		case la_taf_object_spec_dynamic_e:
			la_taf_object_spec_copy(&o, object, "", tas);
			la_resolve_object_f(tas->taf, &o);
			la_array_push(objects, &o.spec.direct);
			break;
		case la_taf_object_spec_referenced_e:
			if (tas->refd_object >= 0) {
				la_array_push(objects, &tas->refd_object);
			} else LA_THROW("there is no referenced object");
			break;
		case la_taf_object_spec_all_held_e:
			la_resolve_all_held_objects(tas, objects);
			break;
	}
	return;
}

