/* common.c - functions for handling stuff in .taf files
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: common.c 402 2004-03-22 11:02:02Z rsnel $
 */
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "../adrift.h"
#include "../verbose.h"
#include "../emem.h"
#include "../char_io.h"
#include "../line_in.h"
#include "../line_out.h"
#include "common.h"

int la_taf_array_total(la_taf_array_t *a) {
	return la_slot_total(a->slots);
}

#define LA_VERBOSE_TYPE Taf_Freer

void la_taf_lines_free(la_taf_lines_t *ls LA_COMMON_DBG_ARG) {
	int i;
	for (i = 0; i < ls->no_lines; i++) free(ls->lines[i]);
	free(ls->lines);
}

#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Parser

void la_taf_lines_parse(la_taf_lines_t *lines, la_line_in_t *lin LA_COMMON_DBG_ARG, int count) {
	char *line;
	int i;
	assert(count >= 0);
	la_array_t a;
	la_array_sinit(&a, (*lines), lines, NULL);
	for (i = 0; i < count; i++) {
		line = la_line_in_get_string(lin);
		LA_DBG2("%s:%s[%d]=\"%s\" (lines)", lin->line_no, name, *a.no_objects, line);
		if (!line) throw();
		la_array_push(&a, &line);
	}
}

void la_taf_postlines_parse(la_taf_lines_t *lines, la_line_in_t *lin LA_COMMON_DBG_ARG, char *term) {
	char *line;
	la_array_t a;
	assert(lines && lin);
	la_array_sinit(&a, (*lines), lines, NULL);
	do { 
		line = la_line_in_get_string(lin);
		LA_DBG2("%s:%s[%d]=\"%s\" (postlines)", lin->line_no, name, *a.no_objects, line);
		if (!line) throw();
		if (!strcmp(line, term)) break;
		la_array_push(&a, &line);
	} while (1);
	free(line);
}

int la_taf_noneg_parse(la_line_in_t *lin LA_COMMON_DBG_ARGS) {
	int in;
	in = la_line_in_get_int(lin);
	LA_DBG2("%s:%s%s=%d (uint)", lin->line_no, prefix, name, in);
	if (in < 0) 
		LA_WARNING("unsigned integer expected on line %s of %s, found %d", lin->line_no, lin->cin->name, in);
	return in;
}

int la_taf_integt_parse(la_line_in_t *lin, int min LA_COMMON_DBG_ARGS) {
	int in;
	in = la_line_in_get_int(lin);
	LA_DBG2("%s:%s%s=%d (int >= %d)", lin->line_no, prefix, name, in, min);
	if (in < min) 
		LA_WARNING("integer >= %d expected on line %s of %s, found %d", lin->line_no, lin->cin->name, in);
	return in;
}

int la_taf_range_parse(la_line_in_t *lin, int min, int max LA_COMMON_DBG_ARGS) {
	int in;
	in = la_line_in_get_int(lin);
	LA_DBG2("%s:%s%s=%d (range from %d to %d)", lin->line_no, prefix, name, in, min, max);
	if (in < min || in > max) 
		LA_WARNING("integer between %d and %d expected on line %s of %s, found %d", min, max, lin->line_no, lin->cin->name, in);
	return in;
}

int la_taf_boolean_parse(la_line_in_t *lin LA_COMMON_DBG_ARGS) {
	int in;
	in = la_line_in_get_int(lin);
	LA_DBG2("%s:%s%s=%d (bool)", lin->line_no, prefix, name, in);
	if (in < 0 || in > 1) 
		LA_WARNING("boolean expected on line %s of %s, found %d", lin->line_no, lin->cin->name, in);
	return in;
}

#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Writer

void la_taf_lines_write(la_taf_lines_t *lines, la_line_out_t *lout LA_COMMON_DBG_ARG) {
	int i;
	for (i = 0; i < lines->no_lines; i++) 
		la_line_out_put_string(lout, lines->lines[i]);
}

void la_taf_postlines_write(la_taf_lines_t *lines, la_line_out_t *lout LA_COMMON_DBG_ARG, char *term) {
	la_taf_lines_write(lines, lout
#ifndef NDEBUG
			, name
#endif
			);
	la_line_out_put_string(lout, term);
}

#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Copier

void la_taf_lines_copy(la_taf_lines_t *out, 
		la_taf_lines_t *in LA_COMMON_DBG_ARG) {
	int i;
	char *line;
	la_array_t a;
	assert(out && in);
	la_array_sinit(&a, (*out), lines, NULL);
	for (i = 0; i < in->no_lines; i++) {
		line = estrdup(in->lines[i]);
		la_array_push(&a, &line);
	}
}
