/* taf.c - functions for handling generic .TAF files
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 * 
 * $Id: taf.c 475 2004-07-18 19:52:33Z rsnel $
 */
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "generic.h"
#include "../verbose.h"
#include "../char_io.h"
#include "../line_in.h"
#include "../line_out.h"
#include "../adrift.h"
#include "../objcache.h"

#define LA_VERBOSE_TYPE Taf_Freer

#ifndef NDEBUG
#define LA_TAF_DEBUG_ARG , "taf."
#else
#define LA_TAF_DEBUG_ARG
#endif

void la_taf_init(la_taf_t *taf) {
	memset(taf, 0, sizeof(*taf));
}

void la_taf_free(la_taf_t *taf) {
	la_tas_t tas;
	tas.taf = taf;
	la_objcache_free(taf);
	switch (taf->version) {
		case 380:
		case 390:
		case 400:
			la_taf_data_free(&taf->data LA_TAF_DEBUG_ARG, &tas);
		case 0:
			break;
		default:
			LA_WARNING("impossibe version number %.02f", 
					taf->version/100.);
			break;
	}
}

#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Copier

void la_taf_copy(la_taf_t *d, la_taf_t *s) {
	la_tas_t tas;
	tas.taf = d;
	d->version = s->version;
	switch (s->version) {
		case 380:
		case 390:
		case 400:
			la_taf_data_copy(&d->data, &s->data LA_TAF_DEBUG_ARG, &tas);
			break;
		case 0:
			break;
		default:
			LA_WARNING("impossibe version number %.02f", 
					s->version/100.);
			break;
	}
	memcpy(d->password, s->password, 8);
}
	
#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Parser

void la_taf_parse_version(la_taf_t *t, la_char_io_t *in, int *obf) {
	char buf[8];
	int xors[8], i;
	double vfloato, vfloatn;
	int vobf = 0, vnobf = 0;
	assert(t && in && current_exception_context);
	assert(*obf == 1);

	t->version = 0;

	LA_DEBUG("parsing version");
	for (i = 0; i < 8; i++) {
		buf[i] = la_char_io_getc(in);
		la_char_filter_command(in, 0, "getXOR", &xors[i]);
	}

	if (strncmp(buf, "Version ", 8)) {
		LA_DEBUG("input in obfuscated mode doesn't seem to be a legal .TAF file...");
		for (i = 0; i < 8; i++) buf[i] ^= xors[i];

		if (strncmp(buf, "Version ", 8)) {
			LA_DEBUG("input in nonobfuscated mode doesn't seem to be a legal .TAF file...");
			LA_WARNING("\"Version \" not found, "
					"probably corrupt/unsupported");
		} else {
			LA_DEBUG("input is probably unobfuscated");
			*obf = 0;
		}
	}

	for (i = 0; i < 4; i++) {
		buf[i] = la_char_io_getc(in);
		la_char_filter_command(in, 0, "getXOR", &xors[i]);
	}
	buf[4] ='\0';

	if (sscanf(buf, "%lf", &vfloato) != 1) {
		LA_DEBUG("input in obfuscated mode doesn't seem to contain a version number");
	} else vobf = 1;
	for (i = 0; i < 4; i++) buf[i] ^= xors[i];
	if (sscanf(buf, "%lf", &vfloatn) != 1) {
		LA_DEBUG("input in nonobfuscated mode doesn't seem to contain a version number");
	} else vnobf = 1;
	
	if (vobf) if (vfloato <= 0 || vfloato >= 10) vobf = 0;
	if (vnobf) if (vfloatn <= 0 || vfloatn >= 10) vnobf = 0;

	if (vobf + vnobf == 2) {
		LA_DEBUG("two legal version numbers (obfuscated and unobfuscated), choosing...");
		if (*obf == 1) t->version = 100*vfloato;
		if (*obf == 0) t->version = 100*vfloatn;
	} else {
		if (vobf == 1) {
			t->version = 100*(vfloato + 0.001);
			if (*obf == 0) LA_WARNING("header heuristics seem wrong");
			*obf = 1;
		} else if (vnobf == 1) {
			t->version = 100*(vfloatn + 0.001);

			if (*obf == 1) LA_WARNING("header heuristics seem wrong");
			*obf = 0;
		}
	}

	if (t->version == 0)
		LA_THROW("unable to read version number, "
				"corrupt/unsupported?");

	LA_VERBOSE(".TAF version is %.02f, in %sobfuscated mode", t->version/100., *obf == 1 ? "" : "un");

	if (*obf == 0) la_char_filter_pop(in);

	la_char_io_read(buf, 2, in);
	
	if (strncmp(buf, "\r\n", 2)) 
		LA_WARNING("version line should end with \"\\r\\n\', "
				"but it doesn't");
}

void la_taf_parse_password(la_taf_t *t, la_char_io_t *in, int *pass) {
	char buf[4];
	int c = EOF;
	assert(t);
	assert(in);
	assert(*pass == 1);
	LA_DEBUG("checking if password exists");
	try c = la_char_io_getc_nowhine(in);
	catch {
		LA_WARNING(".TAF file presumably doesn't have a password, this is nonstandard");
		*pass = 0;
	}
	endtry;

	if (*pass == 0) return;

	*t->password = c;

	LA_DEBUG("parsing password");
	la_char_io_read(t->password + 1, 3, in);

#if 0
	LA_DEBUG("first four characters of password are \"%.*s\"", 
			4, t->password);
#endif
	la_char_io_read(buf, 4, in);
	
	if (strncmp(buf, "Wild", 4)) 
		LA_WARNING("\"Wild\" expected between parts of password, but something else found,\ndata may be corrupt");

	la_char_io_read(t->password + 4, 4, in); 

#if 0
	LA_DEBUG("last four characters of password are \"%.*s\"", 
			4, t->password + 4);
#endif
}

void la_taf_parse_char_io(la_taf_t *taf, la_char_io_t *cin) {
	int c, media_offset, offset, obf = 1, pass = 1; 
	char buf[9];
	la_line_in_t lin;
	la_tas_t tas;
	tas.taf = taf;
	assert(taf && cin && current_exception_context);

	la_char_filter_push(cin, "crypt", NULL);
	la_taf_parse_version(taf, cin, &obf);

	switch (taf->version) {
		case 380:
		case 390:
			la_line_in_init(&lin, cin, 1);
			try {
				la_exception_cleanup_adda(lin, &la_line_in_free);
				la_taf_data_parse(&taf->data, &lin LA_TAF_DEBUG_ARG, &tas);
			} delegate;
			break;
		case 400:
			if (obf) {
				la_char_filter_command(cin, 0, "passive", NULL);
				la_char_io_read(buf, 8, cin);
				buf[8] = '\0';
				if (sscanf(buf, "%d", &media_offset) != 1) 
					LA_THROW("unable to read media offset,"
							"data damaged?");
				LA_DEBUG("media_offset=%d", media_offset);
				la_char_filter_push(cin, "inflate", NULL);
			}
			la_line_in_init(&lin, cin, 1);
			try {
				la_exception_cleanup_adda(lin, &la_line_in_free);
				la_taf_data_parse(&taf->data, &lin LA_TAF_DEBUG_ARG, &tas);
				if (obf) {
					try {
						la_char_io_getc_nowhine(cin);
						LA_WARNING("end of compressed stream expected, but not found");
					}
					endtry;
				}
			} delegate;
			if (obf) {
				la_char_filter_pop(cin);
				if (la_char_io_getc(cin) != 0) LA_WARNING("ASCII NUL byte expected, but other byte found");

				la_char_filter_bytecount(&offset, cin, 0);
				if (media_offset != offset) 
					LA_WARNING("serious inconsistency in .taf file, guessing media offset");

				if (taf->data.datafiles_embedded) 
					la_taf_data_xtra_parse(&taf->data, cin LA_TAF_DEBUG_ARG, &tas);

				la_char_filter_command(cin, 0, "active", NULL);
			} else taf->data.datafiles_embedded = 0;
			break;
		default:
			LA_ERROR("taf version %0.2f unsupported", 
					taf->version/100.);
			taf->version = 0;
			throw();
	}

	la_taf_parse_password(taf, cin, &pass);

	if (pass) {
		if (obf && taf->version == 400) la_char_filter_command(cin, 0, 
				"passive", NULL);

		c = la_char_io_getc(cin);

		if (c != '\r') 
			LA_WARNING("expected character in stream should be "
					"'\\r', but isn't");

		c = la_char_io_getc(cin);

		if (c != '\n') 
			LA_WARNING("expected character in stream should be "
					"'\\n', but isn't");
	
		try {
			la_char_io_getc_nowhine(cin);
			LA_WARNING("junk found at end of file (EOF expected)");
		} endtry;
	}
}

void la_taf_parse_fp(la_taf_t *taf, FILE *fp, char *name) {
	la_char_io_t cin;
	assert(taf && fp && current_exception_context);

	la_char_io_init(&cin, name);

	try {
		la_exception_cleanup_adda(cin, &la_char_io_free);
		la_char_filter_push(&cin, "fpin", fp);
		la_taf_parse_char_io(taf, &cin);
	}
	delegate;
}

void la_taf_parse_file(la_taf_t *taf, char *filename) {
	assert(current_exception_context);
	FILE *fp;

	if (!(fp = fopen(filename, "r"))) 
		LA_THROW("opening %s: %s", filename, strerror(errno));

	try {
		la_exception_cleanup_add(fp, &fclose);
		la_taf_parse_fp(taf, fp, filename);
	}
	delegate;
}

int la_taf_has_password(la_taf_t *taf) {
	return memcmp(taf->password, "        ", 8);
}

int la_taf_validate_password(la_taf_t *taf, char *tovalidate) {
	int len = strlen(tovalidate);
	if (len > 8) len = 8;
	if (!memcmp(tovalidate, taf->password, len)) {
		if (len < 8) if (memcmp(taf->password + len, "       ", 8 - len)) return 0;
		return 1;
	}
	return 0;
}
	
#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Writer

void la_taf_write_version(la_taf_t *t, la_char_io_t *cout) {
	LA_DEBUG("writing version");
	la_char_io_printf(cout, "Version %.02f\r\n", t->version/100.);
}

void la_taf_write_password(la_taf_t *t, la_char_io_t *cout) {
	LA_DEBUG("writing password");
	la_char_io_write(t->password, 4, cout);
	la_char_io_printf(cout, "Wild");
	la_char_io_write(t->password + 4, 4, cout);
}

void write400(la_taf_t *taf, la_char_io_t *cout, int obf, int high) {
	char *bytes = NULL;
	int no_bytes = 0;
	la_char_io_t tmp;
	la_line_out_t lout;
	la_array_t a;
	la_tas_t tas;
	tas.taf = taf;

	if (obf) {
		la_array_ninit(&a, bytes, NULL);
		la_char_io_init(&tmp, "temporary buffer");
		la_line_out_init(&lout, &tmp, 2);
		try {
			la_exception_cleanup_adda(lout, &la_line_out_free);
			la_exception_cleanup_adda(tmp, &la_char_io_free);
			la_exception_cleanup_adda(a, &la_array_free);
			la_char_io_flush(cout);
			la_char_filter_command(cout, 0, "passive", NULL);

			la_char_filter_push(&tmp, "bufout", &a);
			if (high) {
				high = 9;
				la_char_filter_push(&tmp, "deflate", &high);
			} else la_char_filter_push(&tmp, "deflate", NULL);
			la_taf_data_write(&taf->data, &lout LA_TAF_DEBUG_ARG, &tas);
			la_line_out_flush(&lout);
			la_char_io_flush(&tmp);
			la_char_io_printf(cout, "%08d", no_bytes + 23);
			la_char_io_write(bytes, no_bytes, cout);
			la_char_io_putc(0, cout);
			if (taf->data.datafiles_embedded)
				la_taf_data_xtra_write(&taf->data, cout LA_TAF_DEBUG_ARG, &tas);
		}
		delegate;
	} else {
		la_line_out_init(&lout, cout, 2);
		try {
			la_exception_cleanup_adda(lout, &la_line_out_free);
			la_taf_data_write(&taf->data, &lout LA_TAF_DEBUG_ARG, &tas);
			la_line_out_flush(&lout);
		}
		delegate;
		la_char_io_flush(cout);
		if (taf->data.datafiles_embedded)
			LA_WARNING("skipping over embedded data files");
	}
		
	la_char_io_flush(cout);
	if (obf) la_char_filter_command(cout, 0, "active", NULL);
}

void la_taf_write_char_io(la_taf_t *taf, la_char_io_t *cout, int options) {
	int obf = 1, high = 0, pass = 1;
	la_line_out_t lout;
	la_tas_t tas;
	tas.taf = taf;

	if (options&LA_TAF_WUNOBFUSC) {
		LA_VERBOSE("enabling unobfuscated .TAF writing");
		LA_WARNING("unobfuscated .TAF file will not be readable by standard tools");
		obf = 0;
	}

	if (options&LA_TAF_WZLIBHIGH) {
		LA_VERBOSE("enabling high compression for .TAF writing");
		if (!obf && taf->version >= 400) 
			LA_WARNING("high compression cannot be used together with unobfuscated output");
		if (taf->version < 400) 
			LA_WARNING(".TAF files < version 400 don't have compression support");
		high = 1;
	}

	if (options&LA_TAF_WNOPASSWD) {
		LA_VERBOSE("disabling password output for .TAF writing");
		if (obf) LA_WARNING(".TAF files without password will not be readable by standard tools");
		pass = 0;
	}

	if (obf) la_char_filter_push(cout, "crypt", NULL);
	la_taf_write_version(taf, cout);

	if (taf->version == 380 || taf->version == 390) {
		la_line_out_init(&lout, cout, 2);
		try {
			la_exception_cleanup_adda(lout, &la_line_out_free);
			la_taf_data_write(&taf->data, &lout LA_TAF_DEBUG_ARG, &tas);
			la_line_out_flush(&lout);
		}
		delegate;
		la_char_io_flush(cout);
	} else if (taf->version == 400) write400(taf, cout, obf, high);
	else LA_THROW("writing taf version %0.2f unsupported", 
			taf->version/100.);

	if (pass) {
		la_taf_write_password(taf, cout);
		if (taf->version >= 400 && obf) {
			la_char_io_flush(cout);
			la_char_filter_command(cout, 0, "passive", NULL);
		}
		la_char_io_printf(cout, "\r\n");
	}

	la_char_io_flush(cout);
}

void la_taf_write_fp(la_taf_t *taf, FILE *fp, char *name, int options) {
	la_char_io_t cout;
	assert(taf && fp && current_exception_context);

	la_char_io_init(&cout, name);

	try {
		la_exception_cleanup_adda(cout, &la_char_io_free);
		la_char_filter_push(&cout, "fpout", fp);
		la_taf_write_char_io(taf, &cout, options);
	}
	delegate;
}

void la_taf_write_file(la_taf_t *taf, char *filename, int options) {
	assert(current_exception_context);
	FILE *fp;

	if (!(fp = fopen(filename, "w"))) 
		LA_THROW("opening %s: %s", filename, strerror(errno));


	try {
		la_exception_cleanup_add(fp, &fclose);
		la_taf_write_fp(taf, fp, filename, options);
	}
	delegate;
}
