/* tas.c - functions for handling generic .TAS files
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 * 
 * $Id: tas.c 464 2004-04-04 23:13:16Z rsnel $
 */
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "generic.h"
#include "../verbose.h"
#include "../char_io.h"
#include "../line_in.h"
#include "../line_out.h"
#include "../adrift.h"

#define LA_VERBOSE_TYPE Taf_Freer

#ifndef NDEBUG
#define LA_TAF_DEBUG_ARG , "tas."
#else
#define LA_TAF_DEBUG_ARG
#endif

void la_tas_init(la_tas_t *tas, la_taf_t *t, la_char_io_t *green, la_char_io_t *red) {
	memset(tas, 0, sizeof(*tas));
	tas->taf = t;
	tas->green = green;
	tas->red = red;
	tas->refd_character = -1;
	tas->refd_object = -1;
	tas->iam = "I am ";
	tas->i = "I ";
}

void la_tas_free(la_tas_t *tas) {
	assert(tas);
	switch (tas->taf->version) {
		case 380:
		case 390:
		case 400:
			la_tas_data_free(&tas->data LA_TAF_DEBUG_ARG, tas);
		case 0:
			break;
		default:
			LA_WARNING("impossibe version number %.02f", 
					tas->taf->version/100.);
			break;
	}
}

#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Copier

void la_tas_copy(la_tas_t *d, la_tas_t *s) {
	switch (d->taf->version) {
		case 380:
		case 390:
		case 400:
			la_tas_data_copy(&d->data, &s->data LA_TAF_DEBUG_ARG, d);
			break;
		case 0:
			break;
		default:
			LA_WARNING("impossibe version number %.02f", 
					d->taf->version/100.);
			break;
	}
}
	
#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Parser

void la_tas_parse_char_io(la_tas_t *tas, la_char_io_t *cin) {
	la_line_in_t lin;
	assert(tas && cin && current_exception_context);

	la_char_filter_push(cin, "crypt", NULL);

	switch (tas->taf->version) {
		case 380:
		case 390:
		case 400:
			la_line_in_init(&lin, cin, 0);
			try {
				la_exception_cleanup_adda(lin, &la_line_in_free);
				la_tas_data_parse(&tas->data, &lin LA_TAF_DEBUG_ARG, tas);
			} delegate;
			break;
		default:
			LA_ERROR("taf version %0.2f unsupported", 
					tas->taf->version/100.);
			throw();
	}

	try {
		la_char_io_getc_nowhine(cin);
		LA_WARNING("junk found at end of file (EOF expected)");
	} endtry;
}

void la_tas_parse_fp(la_tas_t *tas, FILE *fp, char *name) {
	la_char_io_t cin;
	assert(tas && fp && current_exception_context);

	la_char_io_init(&cin, name);

	try {
		la_exception_cleanup_adda(cin, &la_char_io_free);
		la_char_filter_push(&cin, "fpin", fp);
		la_tas_parse_char_io(tas, &cin);
	}
	delegate;
}

void la_tas_parse_file(la_tas_t *tas, char *filename) {
	assert(current_exception_context);
	FILE *fp;

	if (!(fp = fopen(filename, "r"))) 
		LA_THROW("opening %s: %s", filename, strerror(errno));

	try {
		la_exception_cleanup_add(fp, &fclose);
		la_tas_parse_fp(tas, fp, filename);
	}
	delegate;
}

#undef LA_VERBOSE_TYPE
#define LA_VERBOSE_TYPE Taf_Writer

void la_tas_write_char_io(la_tas_t *tas, la_char_io_t *cout) {
	la_line_out_t lout;

	la_char_filter_push(cout, "crypt", NULL);

	switch (tas->taf->version) {
		case 380:
		case 390:
		case 400:
			la_line_out_init(&lout, cout, 1);
			try {
				la_exception_cleanup_adda(lout, &la_line_out_free);
				la_tas_data_write(&tas->data, &lout LA_TAF_DEBUG_ARG, tas);
				la_line_out_flush(&lout);
			}
			delegate;
			la_char_io_flush(cout);
			break;
		default:
			LA_ERROR("taf version %0.2f unsupported", 
					tas->taf->version/100.);
			throw();
	}
	la_char_io_flush(cout);
}

void la_tas_write_fp(la_tas_t *tas, FILE *fp, char *name) {
	la_char_io_t cout;
	assert(tas && fp && current_exception_context);

	la_char_io_init(&cout, name);

	try {
		la_exception_cleanup_adda(cout, &la_char_io_free);
		la_char_filter_push(&cout, "fpout", fp);
		la_tas_write_char_io(tas, &cout);
	}
	delegate;
}

void la_tas_write_file(la_tas_t *tas, char *filename) {
	assert(current_exception_context);
	FILE *fp;

	if (!(fp = fopen(filename, "w"))) 
		LA_THROW("opening %s: %s", filename, strerror(errno));


	try {
		la_exception_cleanup_add(fp, &fclose);
		la_tas_write_fp(tas, fp, filename);
	}
	delegate;
}
