/* verbose.c - functions for handling verboseness
 *
 * Copyright (C) 2004 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: verbose.c 366 2004-03-14 17:27:24Z rsnel $
 */
#include <stdio.h> 
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h> 
#include <string.h>
#include "verbose.h"
#include "array.h"

la_verbose_channel_t la_verbose_fatal_channel;

la_verbosity_t *v = NULL;

void la_verbose_init_fp_quiet(FILE *fp) {
	assert(!v);
	v = calloc(1, sizeof(*v));
	if (!v) LA_FATAL("out of memory error allocating verbosity struct");

	la_verbose_fp_channel_init(&la_verbose_fatal_channel, fp);

	la_verbose_channel_add_alltypes(Error, &la_verbose_fatal_channel);
}

void la_verbose_init_fp(FILE *fp) {
	la_verbose_init_fp_quiet(fp);
	la_verbose_channel_add_alltypes(Warning, &la_verbose_fatal_channel);
}

void la_verbose_free() {
	assert(v);
	int i, j;

	for (i = 0; i < LA_VERBOSE_NO_TYPES; i++) {
		for (j = 0; j < LA_VERBOSE_NO_LEVELS; j++) 
			free((*v)[i][j].channels);
	}
	free(v);
	v = NULL;
}

void la_verbose_fp_channel_init(la_verbose_channel_t *c, FILE *fp) {
	assert(c);
	assert(fp);
	c->type = FP;
	c->priv = fp;
}

void la_verbose_channel_add(la_verbose_type_t t,
		la_verbose_level_t l, la_verbose_channel_t *c) {
	la_array_t a;
	assert(v);
	assert(c);
	assert(l >= 0 && l < LA_VERBOSE_NO_LEVELS);
	assert(t >= 0 && t < LA_VERBOSE_NO_TYPES);
	la_array_sinit(&a, (*v)[t][l], channels, NULL);
	if (!la_array_push(&a, c)) 
		LA_FATAL("fatal: error adding verbose output: %s\n", a.strerr);
}

void la_verbose_channel_add_alltypes(la_verbose_level_t l, 
		la_verbose_channel_t *c) {
	int i;
	assert(v);
	assert(c);
	assert(l >= 0 && l < LA_VERBOSE_NO_LEVELS);
	for (i = 0; i < LA_VERBOSE_NO_TYPES; i++) 
		la_verbose_channel_add(i, l, c);
}

void la_verbose_channel_remove(la_verbose_type_t t, la_verbose_level_t l, 
		la_verbose_channel_t *c) {
	la_array_t a;
	int i;
	assert(v);
	assert(c);
	assert(l >= 0 && l < LA_VERBOSE_NO_LEVELS);
	assert(t >= 0 && t < LA_VERBOSE_NO_TYPES);
	la_array_sinit(&a, (*v)[t][l], channels, NULL);
	for (i = 0; i < (*v)[t][l].no_channels; i++) 
		if ((*v)[t][l].channels[i].type == c->type &&
				(*v)[t][l].channels[i].priv == c->priv)
			la_array_delete(&a, NULL, i);
}

void la_verbose_voutput(la_verbose_channel_t *c, char *format, va_list ap) {
	assert(c && c->type >= 0 && c->type < LA_VERBOSE_NO_CHANNEL_TYPES);
	switch (c->type) {
		case FP:
			if (c->priv != NULL) /* extra check for FATAL... */
				vfprintf(c->priv, format, ap);
	}
}

void la_verbose_output(la_verbose_channel_t *c, char *format, ...) {
	va_list ap;
	assert(c->type >= 0 && c->type < LA_VERBOSE_NO_CHANNEL_TYPES);
	va_start(ap, format);
	la_verbose_voutput(c, format, ap);
	va_end(ap);
}

char *la_verbose_prefixen[LA_VERBOSE_NO_LEVELS] = {
	"error: ", "warning: ", "", "debug:", "dbg2:", "dbg3:"
};

void la_verbose(la_verbose_type_t t, la_verbose_level_t l, char *format, ...) {
	va_list ap;
	int i;
	assert(l >= 0 && l < LA_VERBOSE_NO_LEVELS);
	assert(t >= 0 && t < LA_VERBOSE_NO_TYPES);
	if (!v) return;
	for (i = 0; i < (*v)[t][l].no_channels; i++) {
		la_verbose_output(&(*v)[t][l].channels[i], "%s",
				la_verbose_prefixen[l]);
		va_start(ap, format);
		la_verbose_voutput(&(*v)[t][l].channels[i], format, ap);
		va_end(ap);
		la_verbose_output(&(*v)[t][l].channels[i], "\n");
	}
}
